from ml_orchestrator.artifacts import Artifact


class Model(Artifact):
    schema_title: str = "system.Model"


class Dataset(Artifact):
    pass


class HTML(Artifact):
    pass


class Markdown(Artifact):
    pass


class Metrics(Artifact):
    def log_metric(self, metric: str, value: float) -> None:
        assert isinstance(value, (int, float, bool, str)), f"Type {type(value)} is not a valid metric value"
        assert isinstance(metric, str), f"Type {type(metric)} is not a valid metric name"
        self.metadata[metric] = value



import dataclasses
from typing import Any, Dict, List, Tuple


@dataclasses.dataclass(unsafe_hash=True)
class EnvironmentParams:
    base_image: str = None
    target_image: str = None
    packages_to_install: List[str] = dataclasses.field(default_factory=lambda: None)
    pip_index_urls: List[str] = dataclasses.field(default_factory=lambda: None)
    output_component_file: str = None
    install_kfp_package: bool = None
    kfp_package_path: str = None

    @classmethod
    def comp_fields(cls) -> Tuple[dataclasses.Field, ...]:
        return dataclasses.fields(cls)

    def comp_vars(self) -> Dict[dataclasses.Field, Any]:
        fields = self.comp_fields()
        ins_vars = dict()
        for field in fields:
            ins_vars[field] = getattr(self, field.name)

        return ins_vars



import dataclasses
from typing import Dict, List, Set, Type

from ml_orchestrator.env_params import EnvironmentParams
from ml_orchestrator.meta_comp import MetaComponent
from ml_orchestrator.utils.field_utils import (
    get_param_meta_data,
    get_param_meta_data_str,
)

IMPORT_COMPOUND = "from kfp.dsl import *\nfrom typing import *\nimport importlib\n"


@dataclasses.dataclass
class ComponentParser:
    def create_function(self, component: MetaComponent) -> str:
        kfp_func_name = component.kfp_func_name
        component_variables = component.comp_vars()
        comp_class = component.__class__
        func_scope = "(\n\t" + ",\n\t".join(self.get_func_params(component_variables)) + "\n)"
        comp_scope = "(\n\t\t" + ",\n\t\t".join(self.get_comp_params(component_variables)) + "\n\t)"
        func_definition = self._format_function_definition(kfp_func_name, func_scope)
        import_compound = self._format_import_compound(comp_class)
        comp_init = f"comp = {comp_class.__name__}{comp_scope}"
        comp_run = "comp.execute()"
        str_func_body = "\n\t".join([import_compound, comp_init, comp_run])
        str_func = func_definition + "\n\t" + str_func_body
        return str_func

    def _format_function_definition(self, kfp_func_name: str, func_scope: str) -> str:
        return f"def {kfp_func_name}{func_scope}:"

    def _format_import_compound(self, comp_class: Type) -> str:
        return f"from {comp_class.__module__} import {comp_class.__name__}"

    @staticmethod
    def create_decorator(env: EnvironmentParams) -> str:
        dec_vars = env.comp_vars()
        prams = ComponentParser.get_func_params(dec_vars, with_typing=False)
        override_params = ComponentParser._get_decorator_override_params(prams)
        dec_scope = "(\n\t" + ",\n\t".join(override_params) + "\n)"
        dec_scope = dec_scope.replace(", '", ", f'").replace("['", "[f'")
        func_definition = f"@component{dec_scope}"
        return func_definition

    @staticmethod
    def _get_decorator_override_params(prams: List[str]) -> List[str]:
        return [p for p in prams if "None" not in p]

    @staticmethod
    def get_func_params(comp_vars: Dict, with_typing: bool = True) -> List[str]:
        return [
            get_param_meta_data_str(*get_param_meta_data(k, v), with_typing=with_typing) for k, v in comp_vars.items()
        ]

    @staticmethod
    def get_comp_params(comp_vars: Dict) -> List[str]:
        return [f"{k.name}={k.name}" for k, v in comp_vars.items()]

    @staticmethod
    def write_to_file(filename: str, content: str) -> None:
        file_content = f"{IMPORT_COMPOUND}\n\n\n{content}"
        file_content = f"# flake8: noqa: F403, F405\n{file_content}"
        with open(filename, "w", encoding="utf-8") as f:
            f.write(file_content)

    def create_kfp_str(self, component: MetaComponent) -> str:
        decorator_str = self.create_decorator(component.env)
        decorator_str = decorator_str.replace(" = ", "=")
        function_str = self.create_function(component)
        kfp_component_str = f"{decorator_str}\n{function_str}"
        kfp_component_str = kfp_component_str.replace("\t", "    ")
        return kfp_component_str + "\n"

    @staticmethod
    def parse_components_to_file(components: List[MetaComponent], filename: str) -> None:
        kfp_str = ""
        for component in components:
            parser = ComponentParser()
            kfp_str += parser.create_kfp_str(component)
            kfp_str += "\n\n"
        all_unique_params_names: Set[str] = set()
        for component in components:
            all_unique_params_names.update([p.name for p in component.comp_fields()])
        comment_str = "\n# " + "\n# ".join([f"{p}: {p}" for p in all_unique_params_names])
        kfp_str = f"{kfp_str}\n{comment_str}"

        ComponentParser.write_to_file(filename, kfp_str)



import abc
import dataclasses
from typing import Any, Dict, Tuple

from ml_orchestrator.env_params import EnvironmentParams


@dataclasses.dataclass
class MetaComponent(abc.ABC):
    @property
    def kfp_func_name(self) -> str:
        return self.__class__.__name__.lower()

    @abc.abstractmethod
    def execute(
        self,
    ) -> None:
        pass

    @classmethod
    def comp_fields(cls) -> Tuple[dataclasses.Field, ...]:
        return dataclasses.fields(cls)

    def comp_vars(self) -> Dict[dataclasses.Field, Any]:
        fields = self.comp_fields()
        ins_vars = dict()
        for field in fields:
            ins_vars[field] = getattr(self, field.name)

        return ins_vars

    @property
    def env(self) -> EnvironmentParams:
        # ml_orchestrator_ver = importlib.metadata.version("ml-orchestrator")
        return EnvironmentParams()



from typing import Dict, Optional


class Artifact:
    schema_title = "system.Artifact"
    schema_version = "0.0.1"

    def __init__(self, name: Optional[str] = None, uri: Optional[str] = None, metadata: Optional[Dict] = None) -> None:
        self.uri = uri or "./"
        self.name = name or "artifact"
        self.metadata = metadata or {}

    @property
    def path(self) -> str:
        return self.uri



from typing import Annotated, TypeVar

from ml_orchestrator.artifacts.artifact import Artifact
from ml_orchestrator.artifacts.artifacts import HTML, Dataset, Markdown, Metrics, Model

_T = TypeVar("_T")
Input = Annotated[_T, "input"]
Output = Annotated[_T, "output"]

__all__ = ["Model", "Dataset", "HTML", "Markdown", "Metrics", "Input", "Output", "Artifact"]



import dataclasses
from typing import Any, Tuple


def get_artifact_type_string(field_type: Any) -> str:
    if "Annotated" in str(field_type):
        artifact_type = repr(field_type.__args__[0]).split("'")[-2].split(".")[-1]

        side = "Input" if "input" in field_type.__metadata__[0] else "Output"
        return f"{side}[{artifact_type}]"
    if not isinstance(field_type, type):
        field_type_name = repr(field_type)
        return field_type_name.replace("typing.", "")

    field_type_name = field_type.__qualname__
    return field_type_name


def get_param_meta_data(field: dataclasses.Field, value: Any) -> Tuple[str, str, Any]:
    name = field.name
    try:
        field_value = value
    except TypeError:
        field_value = None
    if callable(field.default_factory) and field_value is None:
        field_value = field.default_factory()
    field_type = get_artifact_type_string(field.type)
    return name, field_type, field_value


def get_param_meta_data_str(name: str, field_type: str, value: Any, with_typing: bool = True) -> str:
    if field_type == "str":
        value = f'"{value}"'

    if with_typing:
        md_str = f"{name}: {field_type} = {value}"
    else:
        md_str = f"{name} = {value}"

    return md_str



from .artifacts import artifacts
from .comp_parser import ComponentParser
from .env_params import EnvironmentParams
from .meta_comp import MetaComponent

__all__ = ["ComponentParser", "MetaComponent", "EnvironmentParams", "artifacts"]
